#include <ioavr.h>
#include <inavr.h>
#include "uart.h"


#define MAX_PAGE_SIZE 256 //

char fifo_buf[MAX_PAGE_SIZE];

char ch,ch1,ch2,ch3;
unsigned long N_byte;

void port_init(void)
{
  PORTB = 0x07;
  DDRB  = 0x07;
  PORTC = 0x03;
  DDRC  = 0x03;
}

char spi_write_read(char cData)
{
  SPDR = cData;
  while(!(SPSR & (1<<7)));
  return SPDR;
}

void spi_init(void)
{
 SPCR = 0x52; //setup SPI
 SPSR = 0x00; //setup SPI
}

void uart_init(void)
{
 UCSR0B = 0x00; //disable while setting baud rate
 UCSR0A = 0x02;
 UCSR0C = 0x06;
 UBRR0L = 0x01; //set baud rate lo
 UBRR0H = 0x00; //set baud rate hi
 UCSR0B = 0x98;
}

void load_page(char page)
{
}

//        
void program_mode(void)
{
  PORTB &=~0x3C;
  __delay_cycles(300000);
  ch = spi_write_read(0xAC);
  UART_SEND_CHAR(ch);
  ch = spi_write_read(0x53);
  UART_SEND_CHAR(ch);
  ch = spi_write_read(0x00);
  UART_SEND_CHAR(ch);
  ch = spi_write_read(0x00);
  UART_SEND_CHAR(ch);
}

// 3 , 4-
char write_3_read_1(char byte1,char byte2, char byte3)
{
  spi_write_read(byte1);
  spi_write_read(byte2);
  spi_write_read(byte3);
  return spi_write_read(0x00);
}

void write_4(char byte1,char byte2, char byte3, char byte4)
{
  spi_write_read(byte1);
  spi_write_read(byte2);
  spi_write_read(byte3);
  spi_write_read(byte4);
}

//   adr N 
void read_flash(unsigned long adr,unsigned long N)
{
  for (unsigned long i=0;i<N;i++)
  {
    UART_SEND_CHAR(write_3_read_1(0x20,((i+adr)>>8) & 0xFF,(i+adr) & 0xFF));
    UART_SEND_CHAR(write_3_read_1(0x28,((i+adr)>>8) & 0xFF,(i+adr) & 0xFF));
  }
}

void read_EEPROM(unsigned long adr,unsigned long N)
{
  for (unsigned long i=0;i<N;i++)
  {
    UART_SEND_CHAR(write_3_read_1(0xA0,((i+adr)>>8) & 0xFF,(i+adr) & 0xFF));
  }
}

#pragma vector = USART_RX_vect
__interrupt void uart0_rx_isr(void)
{
  //uart has received a character in UDR
  char N;
  char control_summ;
  
  void (*funcptr)(void) = 0x0000;
  
  ch = UART_GET_CHAR();
  switch (ch)
  {
  case 'p': // 
    {
      PORTB = 0x3F;
      DDRB  = 0x3F;
      program_mode();
      break;
    }
  case 't': // 
    {
      for (int i=0; i<64; i++)
        UART_SEND_CHAR(i);
      break;
    }
  case 's': // signature
    { 
      for (int i=0; i<3; i++)
        UART_SEND_CHAR(write_3_read_1(0x30,0x00,i));
      break;
    }
  case 'R': //reset
    {
      funcptr();
    }
  case 'r': //    
    {
      ch  = UART_GET_CHAR();
      ch1 = UART_GET_CHAR();
      N_byte = UART_GET_CHAR();
      N_byte = (N_byte<<8)|UART_GET_CHAR();
      read_flash(ch<<8|ch1,N_byte);
      break;
    }
  case 'q': // EEPROM
    {
      ch  = UART_GET_CHAR();
      ch1 = UART_GET_CHAR();
      N_byte = UART_GET_CHAR();
      N_byte = (N_byte<<8)|UART_GET_CHAR();
      read_EEPROM(ch<<8|ch1,N_byte);
      break;
    }
  case 'f': // 
    {
      UART_SEND_CHAR(write_3_read_1(0x50,0x00,0x00));//fuse bits
      UART_SEND_CHAR(write_3_read_1(0x50,0x08,0x00));//extended fuse bits
      UART_SEND_CHAR(write_3_read_1(0x58,0x00,0x00));//lock bits
      UART_SEND_CHAR(write_3_read_1(0x58,0x08,0x00));//fuse high bits
      break;
    }
  case 'F':  // 
    {
      ch = UART_GET_CHAR();
      ch1 = UART_GET_CHAR();
      ch2 = UART_GET_CHAR();
      ch3 = UART_GET_CHAR();
      write_4(0xAC,0xA0,0x00,ch);//fuse bits
      __delay_cycles(108000);
      write_4(0xAC,0xA4,0x00,ch1);//extended fuse bits
      __delay_cycles(108000);
      write_4(0xAC,0xE0,0x00,ch2);//lock bits
      __delay_cycles(108000);
      write_4(0xAC,0xA8,0x00,ch3);//fuse high bits
      __delay_cycles(108000);
      UART_SEND_CHAR('F');
      break;
    }
  case 'e': // 
    {
      write_4(0xAC,0x80,0x00,0x00);
      __delay_cycles(108000); // 9ms
      UART_SEND_CHAR('e');
      break;
    }
  case 'w': //   
    {
      N = UART_GET_CHAR();
      for (char i=0; i < N; i++)
      {
        fifo_buf[i<<1] = UART_GET_CHAR();
        fifo_buf[(i<<1)+1] = UART_GET_CHAR();
      }
      control_summ = 0;
      for (char i=0; i<N;i++)
      {
        write_4(0x40,0x00,i,fifo_buf[i<<1]);
        control_summ += fifo_buf[i<<1];
        write_4(0x48,0x00,i,fifo_buf[(i<<1)+1]);
        control_summ += fifo_buf[(i<<1)+1];
      }
      UART_SEND_CHAR(control_summ);
      break;
    }
  case 'W': //  
    {
      ch = UART_GET_CHAR();
      ch1 = UART_GET_CHAR();
      write_4(0x4C,ch,ch1,0x00);
      __delay_cycles(54000); // 4.5 ms
      UART_SEND_CHAR('W');
      break;
    }
  case 'y': //  EEPROM
    {
      ch = UART_GET_CHAR();
      ch1 = UART_GET_CHAR();
      ch2 = UART_GET_CHAR();
      write_4(0xC0,ch,ch1,ch2);
      __delay_cycles(108000); // 10 ms
      UART_SEND_CHAR('y');
      break;
    }
  case 'L': // 
    {
      ch = UART_GET_CHAR();
      PORTC &=~ ch;
      break;
    }
  case 'l': // 
    {
      ch = UART_GET_CHAR();
      PORTC |= ch;
      break;
    }
  }
  
}

void init_devices(void)
{
 //stop errant interrupts until set up
 __disable_interrupt(); //disable all interrupts
 port_init();
 uart_init();
 spi_init();
 
 MCUCR = 0x00;
 EICRA = 0x00; //extended ext ints
 EIMSK = 0x00;
 
 TIMSK0 = 0x00; //timer 0 interrupt sources
 TIMSK1 = 0x00; //timer 1 interrupt sources
 TIMSK2 = 0x00; //timer 2 interrupt sources
 
 PCMSK0 = 0x00; //pin change mask 0 
 PCMSK1 = 0x00; //pin change mask 1 
 PCMSK2 = 0x00; //pin change mask 2
 PCICR = 0x00; //pin change enable 
 PRR = 0x00; //power controller
 __enable_interrupt(); //re-enable interrupts
 //all peripherals are now initialized
}

__C_task main( void )
{
  init_devices();
  for(;;)
  {
  };
}
